﻿using System;
using System.Collections.Generic;
using System.Reflection;
using System.Runtime.Serialization;

namespace InterLinq.Types {
	/// <summary>
	/// InterLINQ representation of <see cref="FieldInfo"/>.
	/// </summary>
	/// <seealso cref="InterLinqMemberInfo"/>
	/// <seealso cref="FieldInfo"/>
	[Serializable]
	[DataContract]
	public class InterLinqFieldInfo : InterLinqMemberInfo {
		#region Properties

		private InterLinqType m_fieldType;

		/// <summary>
		/// Gets the <see cref="MemberTypes">MemberType</see>.
		/// </summary>
		/// <seealso cref="FieldInfo.MemberType"/>
		public override MemberTypes MemberType {
			get { return MemberTypes.Field; }
		}

		/// <summary>
		/// Gets or sets the <see cref="InterLinqType"/> of this field.
		/// </summary>
		/// <seealso cref="FieldInfo.FieldType"/>
		[DataMember]
		public InterLinqType FieldType {
			get { return m_fieldType; }
			set { m_fieldType = value; }
		}

		#endregion

		#region Constructors / Initialization

		/// <summary>
		/// Empty constructor.
		/// </summary>
		public InterLinqFieldInfo() {
		}

		/// <summary>
		/// Initializes this class.
		/// </summary>
		/// <param name="fieldInfo">Represented CLR <see cref="FieldInfo"/>.</param>
		public InterLinqFieldInfo(FieldInfo fieldInfo) {
			Initialize(fieldInfo);
		}

		/// <summary>
		/// Initializes this class.
		/// </summary>
		/// <param name="memberInfo">Represented <see cref="MemberInfo"/></param>
		/// <seealso cref="InterLinqMemberInfo.Initialize"/>
		public override void Initialize(MemberInfo memberInfo) {
			base.Initialize(memberInfo);
			var fieldInfo = memberInfo as FieldInfo;
			m_fieldType = InterLinqTypeSystem.Instance.GetInterLinqVersionOf<InterLinqType>(fieldInfo.FieldType);
		}

		#endregion

		#region Methods

		/// <summary>
		/// Returns the CLR <see cref="MemberInfo"/>.
		/// </summary>
		/// <returns>Returns the CLR <see cref="MemberInfo"/>.</returns>
		public override MemberInfo GetClrVersion() {
			InterLinqTypeSystem tsInstance = InterLinqTypeSystem.Instance;
			lock (tsInstance) {
				if (tsInstance.IsInterLinqMemberInfoRegistered(this)) {
					return tsInstance.GetClrVersion<FieldInfo>(this);
				}

				var declaringType = (Type)DeclaringType.GetClrVersion();
				FieldInfo foundField = declaringType.GetField(Name);
				tsInstance.SetClrVersion(this, foundField);
				return foundField;
			}
		}

		/// <summary>
		/// Compares <paramref name="obj"/> to this instance.
		/// </summary>
		/// <param name="obj"><see langword="object"/> to compare.</param>
		/// <returns>True if equal, false if not.</returns>
		public override bool Equals(object obj) {
			if (obj == null || GetType() != obj.GetType()) {
				return false;
			}
			if (!base.Equals(obj)) {
				return false;
			}
			var other = (InterLinqFieldInfo)obj;
			return m_fieldType.Equals(other.m_fieldType);
		}

		/// <summary>
		/// Serves as a hash function for a particular type.
		/// </summary>
		/// <returns>A hash code for the current <see langword="object"/>.</returns>
		public override int GetHashCode() {
			int num = 1708362606;
			num ^= EqualityComparer<InterLinqType>.Default.GetHashCode(m_fieldType);
			return num ^ base.GetHashCode();
		}

		#endregion
	}
}