﻿using System;
using System.Collections.Generic;
using System.Reflection;
using System.Runtime.Serialization;

namespace InterLinq.Types {
	/// <summary>
	/// InterLINQ representation of <see cref="MemberInfo"/>.
	/// </summary>
	/// <seealso cref="MemberInfo"/>
	[Serializable]
	[DataContract]
	public abstract class InterLinqMemberInfo {
		#region Properties

		private InterLinqType m_declaringType;
		private string m_name;

		/// <summary>
		/// Gets the <see cref="MemberTypes">MemberType</see>.
		/// </summary>
		/// <seealso cref="MemberInfo.MemberType"/>
		public abstract MemberTypes MemberType { get; }

		/// <summary>
		/// Gets or sets the name of this <see cref="InterLinqMemberInfo"/>.
		/// </summary>
		/// <seealso cref="MemberInfo.Name"/>
		[DataMember]
		public string Name {
			get { return m_name; }
			set { m_name = value; }
		}

		/// <summary>
		/// Gets or sets the class that declares this <see cref="InterLinqMemberInfo"/>.
		/// </summary>
		/// <seealso cref="MemberInfo.DeclaringType"/>
		[DataMember]
		public InterLinqType DeclaringType {
			get { return m_declaringType; }
			set { m_declaringType = value; }
		}

		#endregion

		#region Constructors / Initialization

		/// <summary>
		/// Empty constructor.
		/// </summary>
		public InterLinqMemberInfo() {
		}

		/// <summary>
		/// Initializes this class.
		/// </summary>
		/// <param name="memberInfo">Represented CLR <see cref="MemberInfo"/>.</param>
		public InterLinqMemberInfo(MemberInfo memberInfo) {
			Initialize(memberInfo);
		}

		/// <summary>
		/// Initializes this class.
		/// </summary>
		/// <param name="memberInfo">Represented <see cref="MemberInfo"/></param>
		public virtual void Initialize(MemberInfo memberInfo) {
			m_name = memberInfo.Name;
			m_declaringType = InterLinqTypeSystem.Instance.GetInterLinqVersionOf<InterLinqType>(memberInfo.DeclaringType);
		}

		#endregion

		#region Methods

		/// <summary>
		/// Returns the CLR <see cref="MemberInfo"/>.
		/// </summary>
		/// <returns>Returns the CLR <see cref="MemberInfo"/>.</returns>
		public abstract MemberInfo GetClrVersion();

		/// <summary>
		/// Compares <paramref name="obj"/> to this instance.
		/// </summary>
		/// <param name="obj"><see langword="object"/> to compare.</param>
		/// <returns>True if equal, false if not.</returns>
		public override bool Equals(object obj) {
			if (obj == null || GetType() != obj.GetType()) {
				return false;
			}
			var other = (InterLinqMemberInfo)obj;
			return MemberType == other.MemberType && m_name == other.m_name && m_declaringType.Equals(other.m_declaringType);
		}

		/// <summary>
		/// Serves as a hash function for a particular type.
		/// </summary>
		/// <returns>A hash code for the current <see langword="object"/>.</returns>
		public override int GetHashCode() {
			int num = -657106008;
			num ^= EqualityComparer<string>.Default.GetHashCode(m_name);
			num ^= EqualityComparer<InterLinqType>.Default.GetHashCode(m_declaringType);
			return num;
		}

		#endregion
	}
}