﻿using System;
using System.Collections.Generic;
using System.Reflection;
using System.Runtime.Serialization;

namespace InterLinq.Types {
	/// <summary>
	/// InterLINQ representation of <see cref="MethodBase"/>.
	/// </summary>
	/// <seealso cref="InterLinqMemberInfo"/>
	/// <seealso cref="MethodBase"/>
	[Serializable]
	[DataContract]
	public abstract class InterLinqMethodBase : InterLinqMemberInfo {
		#region Properties

		private List<InterLinqType> m_parameterTypes = new List<InterLinqType>();

		/// <summary>
		/// Gets or sets the ParameterTypes.
		/// </summary>
		[DataMember]
		public List<InterLinqType> ParameterTypes {
			get { return m_parameterTypes; }
			set { m_parameterTypes = value; }
		}

		#endregion

		#region Constructors / Initialization

		/// <summary>
		/// Empty constructor.
		/// </summary>
		public InterLinqMethodBase() {
		}

		/// <summary>
		/// Initializes this class.
		/// </summary>
		/// <param name="methodBase">Represented CLR <see cref="MethodBase"/>.</param>
		public InterLinqMethodBase(MethodBase methodBase) {
			Initialize(methodBase);
		}

		/// <summary>
		/// Initializes this class.
		/// </summary>
		/// <param name="memberInfo">Represented <see cref="MemberInfo"/></param>
		/// <seealso cref="InterLinqMemberInfo.Initialize"/>
		public override void Initialize(MemberInfo memberInfo) {
			base.Initialize(memberInfo);
			var methodBase = memberInfo as MethodBase;
			foreach (ParameterInfo parameter in methodBase.GetParameters()) {
				m_parameterTypes.Add(InterLinqTypeSystem.Instance.GetInterLinqVersionOf<InterLinqType>(parameter.ParameterType));
			}
		}

		#endregion

		#region Methods

		/// <summary>
		/// Compares <paramref name="obj"/> to this instance.
		/// </summary>
		/// <param name="obj"><see langword="object"/> to compare.</param>
		/// <returns>True if equal, false if not.</returns>
		public override bool Equals(object obj) {
			if (obj == null || GetType() != obj.GetType()) {
				return false;
			}
			var other = (InterLinqMethodBase)obj;
			if (m_parameterTypes.Count != other.m_parameterTypes.Count) {
				return false;
			}

			for (int i = 0; i < m_parameterTypes.Count; i++) {
				if (m_parameterTypes[i].Equals(other.m_parameterTypes[i])) {
					return false;
				}
			}

			return base.Equals(obj);
		}

		/// <summary>
		/// Serves as a hash function for a particular type.
		/// </summary>
		/// <returns>A hash code for the current <see langword="object"/>.</returns>
		public override int GetHashCode() {
			int num = -103514268;
			num ^= EqualityComparer<List<InterLinqType>>.Default.GetHashCode(m_parameterTypes);
			return num ^ base.GetHashCode();
		}

		#endregion
	}
}