﻿using System;
using System.Collections.Generic;
using System.Reflection;
using System.Runtime.Serialization;

namespace InterLinq.Types {
	/// <summary>
	/// InterLINQ representation of <see cref="PropertyInfo"/>.
	/// </summary>
	/// <seealso cref="InterLinqMemberInfo"/>
	/// <seealso cref="PropertyInfo"/>
	[Serializable]
	[DataContract]
	public class InterLinqPropertyInfo : InterLinqMemberInfo {
		#region Properties

		private InterLinqType m_propertyType;

		/// <summary>
		/// Gets the <see cref="MemberTypes">MemberType</see>.
		/// </summary>
		/// <seealso cref="InterLinqMemberInfo.MemberType"/>
		/// <seealso cref="PropertyInfo.MemberType"/>
		public override MemberTypes MemberType {
			get { return MemberTypes.Property; }
		}

		/// <summary>
		/// Gets or sets the <see cref="InterLinqType"/> of this property.
		/// </summary>
		/// <seealso cref="PropertyInfo.PropertyType"/>
		[DataMember]
		public InterLinqType PropertyType {
			get { return m_propertyType; }
			set { m_propertyType = value; }
		}

		#endregion

		#region Constructors / Initialization

		/// <summary>
		/// Empty constructor.
		/// </summary>
		public InterLinqPropertyInfo() {
		}

		/// <summary>
		/// Initializes this class.
		/// </summary>
		/// <param name="fieldInfo">Represented CLR <see cref="PropertyInfo"/>.</param>
		public InterLinqPropertyInfo(PropertyInfo fieldInfo) {
			Initialize(fieldInfo);
		}

		/// <summary>
		/// Initializes this class.
		/// </summary>
		/// <param name="memberInfo">Represented <see cref="MemberInfo"/></param>
		/// <seealso cref="InterLinqMemberInfo.Initialize"/>
		public override void Initialize(MemberInfo memberInfo) {
			base.Initialize(memberInfo);
			var propertyInfo = memberInfo as PropertyInfo;
			m_propertyType = InterLinqTypeSystem.Instance.GetInterLinqVersionOf<InterLinqType>(propertyInfo.PropertyType);
		}

		#endregion

		#region Methods

		/// <summary>
		/// Returns the CLR <see cref="MemberInfo"/>.
		/// </summary>
		/// <returns>Returns the CLR <see cref="MemberInfo"/>.</returns>
		public override MemberInfo GetClrVersion() {
			InterLinqTypeSystem tsInstance = InterLinqTypeSystem.Instance;
			lock (tsInstance) {
				if (tsInstance.IsInterLinqMemberInfoRegistered(this)) {
					return tsInstance.GetClrVersion<PropertyInfo>(this);
				}

				var declaringType = (Type)DeclaringType.GetClrVersion();
				PropertyInfo foundProperty = declaringType.GetProperty(Name);
				tsInstance.SetClrVersion(this, foundProperty);
				return foundProperty;
			}
		}

		/// <summary>
		/// Compares <paramref name="obj"/> to this instance.
		/// </summary>
		/// <param name="obj"><see langword="object"/> to compare.</param>
		/// <returns>True if equal, false if not.</returns>
		public override bool Equals(object obj) {
			if (obj == null || GetType() != obj.GetType()) {
				return false;
			}
			if (!base.Equals(obj)) {
				return false;
			}
			var other = (InterLinqPropertyInfo)obj;
			return m_propertyType.Equals(other.m_propertyType);
		}

		/// <summary>
		/// Serves as a hash function for a particular type.
		/// </summary>
		/// <returns>A hash code for the current <see langword="object"/>.</returns>
		public override int GetHashCode() {
			int num = -1141188190;
			num ^= EqualityComparer<InterLinqType>.Default.GetHashCode(m_propertyType);
			return num ^ base.GetHashCode();
		}

		#endregion
	}
}