﻿using System;
using System.Collections.Generic;
using System.Data.Linq;
using System.Linq;
using System.Reflection;
using System.Threading;
using InterLinq.Base;

namespace InterLinq.Objects {
	/// <summary>
	/// LINQ to Objects specific implementation of the
	/// <see cref="IQueryHandler"/>.
	/// </summary>
	/// <seealso cref="IQueryHandler"/>
	public class LinqToSQLQueryHandler : IQueryHandler, IBackendProvider {
		#region Fields

		private readonly DataContext m_dataset;

		#endregion

		#region Constructors

		/// <summary>
		/// Initializes this class.
		/// </summary>
		/// <param name="objectSource"><see cref="IObjectSource"/> used get all objects of a type.</param>
		public LinqToSQLQueryHandler(DataContext objectSource) {
			if (objectSource == null) {
				throw new ArgumentNullException("objectSource");
			}
			m_dataset = objectSource;
		}

		#endregion

		private readonly List<Type> m_Types = new List<Type>();

		#region IBackendProvider Members

		public object BeginTransaction() {
			//no explicit transaction, we use implicit transacions.
			return null;
		}

		public void CommitTransaction(object transaction) {
			m_dataset.SubmitChanges();
		}

		public void SaveChanges(IList<BackendProviderNewSet> Changes) {
			foreach (BackendProviderNewSet change in Changes) {
				m_dataset.GetTable(change.type).InsertOnSubmit(change.obj);
			}
			//submit change called on commit transaction
			return;
		}

		public void SaveChanges(IList<BackendProviderChangeSet> Changes) {
			//submit change called on commit transaction
			return;
		}

		public void SaveChanges(IList<BackendProviderDeleteSet> Changes) {
			foreach (BackendProviderDeleteSet change in Changes) {
				m_dataset.GetTable(change.type).DeleteOnSubmit(change.obj);
			}
			//submit change called on commit transaction
			return;
		}

		#endregion

		#region IQueryHandler Members

		public ReaderWriterLock LockOfDataSource {
			get { return null; }
		}

		public IEnumerable<Type> GetTableTypes() {
			return m_Types;
		}

		/// <summary>
		/// Returns an <see cref="IQueryable{T}"/>.
		/// </summary>
		/// <typeparam name="T">Generic Argument of the returned <see cref="IQueryable{T}"/>.</typeparam>
		/// <returns>Returns an <see cref="IQueryable{T}"/>.</returns>
		/// <seealso cref="IQueryHandler.GetTable{T}"/>
		public IQueryable<T> GetTable<T>() where T : class {
			if (!m_Types.Contains(typeof(T))) {
				m_Types.Add(typeof(T));
			}
			return m_dataset.GetTable<T>().AsQueryable();
		}

		/// <summary>
		/// Returns an <see cref="IQueryable{T}"/>.
		/// </summary>
		/// <param name="type">Type of the returned <see cref="IQueryable{T}"/>.</param>
		/// <returns>Returns an <see cref="IQueryable{T}"/>.</returns>
		/// <seealso cref="IQueryHandler.GetTable"/>
		public IQueryable GetTable(Type type) {
			if (!m_Types.Contains(type)) {
				m_Types.Add(type);
			}
			MethodInfo getTableMethod = GetType().GetMethod("GetTable", BindingFlags.Instance | BindingFlags.Public, null,
															  new Type[0], null);
			MethodInfo genericGetTableMethod = getTableMethod.MakeGenericMethod(type);
			return (IQueryable)genericGetTableMethod.Invoke(this, new object[0]);
		}

		/// <summary>
		/// Tells the <see cref="IQueryHandler"/> to start a new the session.
		/// </summary>
		/// <returns>True, if the session creation was successful. False, if not.</returns>
		/// <seealso cref="IQueryHandler.StartSession"/>
		public bool StartSession() {
			return true;
		}

		/// <summary>
		/// Tells the <see cref="IQueryHandler"/> to close the current session.
		/// </summary>
		/// <returns>True, if the session closing was successful. False, if not.</returns>
		/// <seealso cref="IQueryHandler.CloseSession"/>
		public bool CloseSession() {
			return true;
		}

		#endregion
	}
}