﻿using System;
using System.Collections.ObjectModel;
using System.Linq.Expressions;
using System.Runtime.Serialization;
using System.Text;
using InterLinq.Expressions.Helpers;
using InterLinq.Expressions.SerializableTypes;

namespace InterLinq.Expressions {
	/// <summary>
	/// A serializable version of <see cref="MemberInitExpression"/>.
	/// </summary>
	[Serializable]
	[DataContract]
	public class SerializableMemberInitExpression : SerializableExpression {
		#region Constructors

		/// <summary>
		/// Default constructor for serialization.
		/// </summary>
		public SerializableMemberInitExpression() {
		}

		/// <summary>
		/// Constructor with an <see cref="MemberInitExpression"/> and an <see cref="ExpressionConverter"/>.
		/// </summary>
		/// <param name="expression">The original, not serializable <see cref="Expression"/>.</param>
		/// <param name="expConverter">The <see cref="ExpressionConverter"/> to convert contained <see cref="Expression">Expressions</see>.</param>
		public SerializableMemberInitExpression(MemberInitExpression expression, ExpressionConverter expConverter)
			: base(expression, expConverter) {
			m_newExpression = expression.NewExpression.MakeSerializable<SerializableNewExpression>(expConverter);
			m_bindings = expConverter.ConvertToSerializableObjectCollection<SerializableMemberBinding>(expression.Bindings);
		}

		#endregion

		#region Properties

		private ReadOnlyCollection<SerializableMemberBinding> m_bindings;
		private SerializableNewExpression m_newExpression;

		/// <summary>
		/// See <see cref="MemberInitExpression.NewExpression"/>
		/// </summary>
		[DataMember]
		public SerializableNewExpression NewExpression {
			get { return m_newExpression; }
			set { m_newExpression = value; }
		}

		/// <summary>
		/// See <see cref="MemberInitExpression.Bindings"/>
		/// </summary>
		[DataMember]
		public ReadOnlyCollection<SerializableMemberBinding> Bindings {
			get { return m_bindings; }
			set { m_bindings = value; }
		}

		#endregion

		#region ToString() Methods

		/// <summary>
		/// Builds a <see langword="string"/> representing the <see cref="Expression"/>.
		/// </summary>
		/// <param name="builder">A <see cref="System.Text.StringBuilder"/> to add the created <see langword="string"/>.</param>
		internal override void BuildString(StringBuilder builder) {
			if (builder == null) {
				throw new ArgumentNullException("builder");
			}
			if ((m_newExpression.Arguments.Count == 0) && m_newExpression.Type.Name.Contains("<")) {
				builder.Append("new");
			}
			else {
				m_newExpression.BuildString(builder);
			}
			builder.Append(" {");
			int num = 0;
			int count = m_bindings.Count;
			while (num < count) {
				SerializableMemberBinding binding = m_bindings[num];
				if (num > 0) {
					builder.Append(", ");
				}
				binding.BuildString(builder);
				num++;
			}
			builder.Append("}");
		}

		#endregion
	}
}