﻿using System;
using System.Collections.ObjectModel;
using System.Linq.Expressions;
using System.Reflection;
using System.Runtime.Serialization;
using System.Text;
using InterLinq.Expressions.Helpers;
using InterLinq.Types;

namespace InterLinq.Expressions {
	/// <summary>
	/// A serializable version of <see cref="NewArrayExpression"/>.
	/// </summary>
	[Serializable]
	[DataContract]
	public class SerializableNewExpression : SerializableExpression {
		#region Constructors

		/// <summary>
		/// Default constructor for serialization.
		/// </summary>
		public SerializableNewExpression() {
		}

		/// <summary>
		/// Constructor with an <see cref="NewExpression"/> and an <see cref="ExpressionConverter"/>.
		/// </summary>
		/// <param name="expression">The original, not serializable <see cref="Expression"/>.</param>
		/// <param name="expConverter">The <see cref="ExpressionConverter"/> to convert contained <see cref="Expression">Expressions</see>.</param>
		public SerializableNewExpression(NewExpression expression, ExpressionConverter expConverter)
			: base(expression, expConverter) {
			m_arguments = expression.Arguments.MakeSerializableCollection<SerializableExpression>(expConverter);
			m_members = InterLinqTypeSystem.Instance.GetCollectionOf<InterLinqMemberInfo>(expression.Members);
			m_constructor = InterLinqTypeSystem.Instance.GetInterLinqVersionOf<InterLinqConstructorInfo>(expression.Constructor);
		}

		#endregion

		#region Properties

		private ReadOnlyCollection<SerializableExpression> m_arguments;
		private InterLinqConstructorInfo m_constructor;

		private ReadOnlyCollection<InterLinqMemberInfo> m_members;

		/// <summary>
		/// See <see cref="NewExpression.Arguments"/>
		/// </summary>
		[DataMember]
		public ReadOnlyCollection<SerializableExpression> Arguments {
			get { return m_arguments; }
			set { m_arguments = value; }
		}

		/// <summary>
		/// See <see cref="NewExpression.Members"/>
		/// </summary>
		[DataMember]
		public ReadOnlyCollection<InterLinqMemberInfo> Members {
			get { return m_members; }
			set { m_members = value; }
		}

		/// <summary>
		/// See <see cref="NewExpression.Constructor"/>
		/// </summary>
		[DataMember]
		public InterLinqConstructorInfo Constructor {
			get { return m_constructor; }
			set { m_constructor = value; }
		}

		#endregion

		#region ToString() Methods

		internal override void BuildString(StringBuilder builder) {
			if (builder == null) {
				throw new ArgumentNullException("builder");
			}
			InterLinqType type;
			type = (m_constructor == null) ? (type = base.Type) : m_constructor.DeclaringType;
			builder.Append("new ");
			int count = m_arguments.Count;
			builder.Append(type.Name);
			builder.Append("(");
			if (count > 0) {
				for (int i = 0; i < count; i++) {
					if (i > 0) {
						builder.Append(", ");
					}
					if (m_members != null) {
						string propertyName = null;
						if ((m_members[i].MemberType == MemberTypes.Method) &&
							((propertyName = GetPropertyNoThrow((InterLinqMethodInfo)m_members[i])) != null)) {
							builder.Append(propertyName);
						}
						else {
							builder.Append(m_members[i].Name);
						}
						builder.Append(" = ");
					}
					m_arguments[i].BuildString(builder);
				}
			}
			builder.Append(")");
		}

		private static string GetPropertyNoThrow(InterLinqMethodInfo method) {
			if (method != null) {
#warning Implement
				//InterLinqType declaringType = method.DeclaringType;
				//BindingFlags bindingAttr = BindingFlags.NonPublic | BindingFlags.Public;
				//bindingAttr |= method.IsStatic ? BindingFlags.Static : BindingFlags.Instance;
				//foreach( PropertyInfo info in declaringType.GetProperties( bindingAttr ) ) {
				//    if( info.CanRead && ( method == info.GetGetMethod( true ) ) ) {
				//        return info;
				//    }
				//    if( info.CanWrite && ( method == info.GetSetMethod( true ) ) ) {
				//        return info;
				//    }
				//}
			}
			return null;
		}

		#endregion
	}
}