﻿#region Header

/*
Remote views based on linq
By

Pascal Hauser 
Dipl. Ing. in Informatik, Hochschule für Technik Rapperswil, 2006
Master Thesis, Hochschule für Technik Rapperswil, 2008-2010

and

Raphael Gfeller
Dipl. Ing. in Informatik, Hochschule für Technik Rapperswil, 2006
Master Thesis, Hochschule für Technik Rapperswil, 2008-2010

*/

#endregion

#region Usings

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using InterLinq.Client;
using InterLinq.Objects;
using InterLinq.Sebarf;
using InterLinq.Server.Services.Impl;
using InterLinq.Services.Interfaces;
using InterLinq.Test.DH;
using InterLinq.Test.PD;
using InterLinq.Test.UI;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Sebarf.Client;
using Sebarf.Configuration;
using Sebarf.Core;
using Sebarf.Core.Internal;
using Sebarf.Diagnostics.Interfaces;
using Sebarf.ProcessChains.Interfaces;
using Sebarf.Profiles.WPF;
using Sebarf.Server;
using Sebarf.Services.Impl;
using Sebarf.Services.Interfaces;
using Sebarf.Session.Interfaces;
using Sebastof.Server.Services.Impl;

#endregion

namespace InterLinq.Test {
	/// <summary>
	/// realizes a application that is able to query for objects
	/// </summary>
	[TestClass]
	public class TestSampleApplication {
		#region Public Methods

		[TestMethod]
		[Description("starts a test application that is able to query for objects")]
		[DeploymentItem("TestSampleApplication.RequestHandler.Client.Internal.xaml")]
		[DeploymentItem("TestSampleApplication.RequestHandler.Client.Application.xaml")]
		[DeploymentItem("TestSampleApplication.RequestHandler.Server.Internal.xaml")]
		[DeploymentItem("TestSampleApplication.RequestHandler.Server.Application.xaml")]
		public void TestSampleApplicationUI() {
			// run app in a wpf context
			WPFTestHelper.RunTest(RunSampleApplication);
		}

		/// <summary>
		/// method handle
		/// </summary>
		public static void RunSampleApplication() {
			// create sample data
			Dictionary<Type, IList> data = SampleDataFactory.CreateInterLinqOrdersWithContent(10000, 20000);

			// remeber the products on the server
			IQueryable<IProduct> serverProducts = ((List<IProduct>)data[typeof(IProduct)]).AsQueryable();

			// create server query handler
			var queryHandler = new ObjectQueryHandler(new ObjectBasedDataSource(data));

			// create server
			IServer server = WcfBasedServer.Create(delegate {
				// create the base service
				var locator = new ServiceRegistrationLocatorService();
				// register backend locator service
				locator.RegisterService(new BackendProviderService());
				// register query handler locator service
				locator.RegisterService(new QueryHandlerLocatorService { QueryHandler = queryHandler });
				// create engine
				RequestHandlerEngine toReturn = RequestHandlerEngine.Create(locator,
																			 RequestHandlerLoader
																				.Load(
																				"TestSampleApplication.RequestHandler.Server.Application.xaml"),
																			 RequestHandlerLoader
																				.Load(
																				"TestSampleApplication.RequestHandler.Server.Internal.xaml"));
				// add interlinq handler
				locator.GetRegisteredService<IAddRequestHandlerService>().
					AddAddionalHandler(new ExecuteLinqQueryRequestHandler { QueryHandler = queryHandler });
				// set backend provider
				locator.GetRegisteredService<IBackendProviderRegisterService>().
					SetBackendProvider(queryHandler);

				return toReturn;
			});

			try {
				// starts the interlinq server
				server.Start();

				// create client

				// create base service
				var clientServiceLocator = new ServiceRegistrationLocatorService();

				// create engine
				IClient client = WcfBasedClient.Create("Client 1",
														RequestHandlerEngine.Create(clientServiceLocator,
																					 RequestHandlerLoader.Load(
																						"TestSampleApplication.RequestHandler.Client.Internal.xaml")));
				// connect
				client.Connect();

				// load profiles
				client.Process(new WpfBasedProfileLoadRequest("Profile", "CacheBasedLazyLoadProfile_Server.xaml"));
				client.ProcessLocal(new WpfBasedProfileLoadRequest("Profile", "CacheBasedLazyLoadProfile.xaml"));

				// populate remote engine
				clientServiceLocator.GetRegisteredService<IRequestRemoteHandlerEngineLocatorService>().SetRemoteEngine(client);

				// create client datacontext
				var dataContext1 =
					new SampleDataContext(
						new InterLinqQueryHandler(
							new InterLinqQueryProvider(new SebarfBasedLinqQueryGateway(clientServiceLocator, client))));
				clientServiceLocator.RegisterService(new QueryHandlerLocatorService { QueryHandler = dataContext1.QueryHandler });

				// ensure dependencies
				client.Process(new ResolveServicesDependenciesRequest());
				client.Process(new ResolveProcessChainDependenciesRequest());
				client.ProcessLocal(new ResolveServicesDependenciesRequest());
				client.ProcessLocal(new ResolveProcessChainDependenciesRequest());

				// startup services
				client.Process(new SetupServicesRequest());
				client.ProcessLocal(new SetupServicesRequest());

				// create and view the ui which displayed the generated products
				new SampleApplication(500,
									   new SampleApplication.ContentEntry {
										   UserControl = new ProductsView(dataContext1.Products, clientServiceLocator),
										   UserControlName = "Products.Client"
									   },
									   new SampleApplication.ContentEntry { UserControl = new ProductsView(serverProducts, null), UserControlName = "Products.Server" },
									   new SampleApplication.ContentEntry {
										   UserControl = new ServicesStatisticsView(clientServiceLocator),
										   UserControlName = "Services.Statistics"
									   },
									   new SampleApplication.ContentEntry { UserControl = new ServicesView(clientServiceLocator), UserControlName = "Services" }).
					ShowDialog();

				// close session
				client.Process(new CloseSessionRequest());
			}
			catch (Exception ex) {
				// error handling
				Logger.WriteDebug(ex.StackTrace);
				Assert.Fail(ex.Message);
			}
			finally {
				// stop the server
				server.Stop();
			}
		}

		#endregion
	}
}