#region Header

/*
Remote views based on linq
By

Pascal Hauser 
Dipl. Ing. in Informatik, Hochschule fr Technik Rapperswil, 2006
Master Thesis, Hochschule fr Technik Rapperswil, 2008-2010

and

Raphael Gfeller
Dipl. Ing. in Informatik, Hochschule fr Technik Rapperswil, 2006
Master Thesis, Hochschule fr Technik Rapperswil, 2008-2010

*/

#endregion

#region Usings

using System;
using System.Runtime.InteropServices;

#endregion

namespace Utils {
	/// <summary>
	/// Represents a time span. It is used to convert ticks to milliseconds.
	/// </summary>
	[Serializable]
	public class HighResolutionTimeSpan {
		private readonly double m_Frequency;
		private readonly double m_Miliseconds;

		/// <summary>
		/// Constructor
		/// </summary>
		public HighResolutionTimeSpan() {
			QueryPerformanceFrequency(ref m_Frequency);
		}

		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="Ticks"></param>
		public HighResolutionTimeSpan(double Ticks) {
			QueryPerformanceFrequency(ref m_Frequency);

			m_Miliseconds = 1000 * (Ticks / m_Frequency);
		}

		/// <summary>
		/// Gets the frequenzy of the performance timer
		/// </summary>
		public double Frequency {
			get { return m_Frequency; }
		}

		/// <summary>
		/// Gets number of miliseconds
		/// </summary>
		public double Miliseconds {
			get { return m_Miliseconds; }
		}

		[DllImport("KERNEL32")]
		private static extern bool QueryPerformanceFrequency(ref double lpFrequency);

		/// <summary>
		/// Returns number of milliseconds as string.
		/// </summary>
		/// <returns></returns>
		public override string ToString() {
			return String.Format("{0}", m_Miliseconds);
		}
	}

	/// <summary>
	/// This class is used to meassure time between the calls to Start and Stop.
	/// </summary>
	/// <example>
	/// <code>
	/// PerformanceTimer pt = new PerformanceTimer();
	/// pt.Start();
	/// //do something
	/// pt.Stop();
	/// Console.WriteLine(pt.ElapsedTimeMilliseconds);
	/// </code>
	/// </example>
	[Serializable]
	public class PerformanceTimer {
		private readonly double m_Frequency;
		private double m_ElapsedCount;
		private double m_StartCount;
		private double m_StopCount;

		/// <summary>
		/// Creates an Instance of the <see cref="PerformanceTimer"/> and initializes 
		/// the <see cref="Frequency"/> of the high-resolution performance counter.
		/// </summary>
		public PerformanceTimer() {
			m_Frequency = 0;
			QueryPerformanceFrequency(ref m_Frequency);
		}

		/// <summary>
		/// Returns the elapsed counts of the last performance measure.
		/// </summary>
		public double Elapsed {
			get { return m_ElapsedCount; }
		}

		/// <summary>
		/// Returns the elapsed seconds of the last performance measure.
		/// </summary>
		public double Seconds {
			get { return (m_ElapsedCount / m_Frequency); }
		}

		/// <summary>
		/// Returns the Frequency of of the high-resolution performance counter.
		/// </summary>
		public double Frequency {
			get { return m_Frequency; }
		}

		/// <summary>
		/// Returns the elapsed time in milliseconds of the last performance measure.
		/// </summary>
		public HighResolutionTimeSpan ElapsedTimeMilliseconds {
			get { return new HighResolutionTimeSpan(m_ElapsedCount); }
		}

		[DllImport("KERNEL32")]
		private static extern bool QueryPerformanceCounter(ref double lpPerformanceCount);

		[DllImport("KERNEL32")]
		private static extern bool QueryPerformanceFrequency(ref double lpFrequency);

		/// <summary>
		/// Starts the performance measuring.
		/// </summary>
		public void Start() {
			m_StartCount = 0;
			QueryPerformanceCounter(ref m_StartCount);
		}

		/// <summary>
		/// Stops the performance measuring.
		/// </summary>
		public void Stop() {
			m_StopCount = 0;
			QueryPerformanceCounter(ref m_StopCount);
			m_ElapsedCount = m_StopCount - m_StartCount;
		}

		/// <summary>
		/// Resets all internal variables.
		/// </summary>
		public void Reset() {
			m_StartCount = 0;
			m_StopCount = 0;
			m_ElapsedCount = 0;
		}
	}
}